from ._time_control import time_control


class time_compare(time_control):
    """Class for getting the intersection of two time axis.

    The :class:`time_compare` contains the class `time_control`.

    Parameters
    ----------
    compare_objects: str or list or nested list
        List of all objects to compare their time axes
    """

    def __init__(self, *compare_objects, **kwargs):
        self.compare_objects = self.compare_objects(compare_objects)
        self.time_control_objects = self.time_control_objects(**kwargs)
        self.times = self.times()

    def compare_objects(self, compare_objects):
        """List of all objects to compare their time axes."""
        return self._flatten_list(compare_objects)

    def time_control_objects(self, **kwargs):
        """List of all `compare_objects` set to time_control objects."""
        return [
            self._to_time_control_object(
                cpo,
                **kwargs,
            )
            for cpo in self.compare_objects
        ]

    def times(self):
        """List of all time axes read from ``time_control_objects``."""
        return [tco.time for tco in self.time_control_objects]

    def _to_time_control_object(self, identity, **kwargs):
        """Open identity as ``time_control`` object."""
        if isinstance(identity, time_control):
            return identity
        else:
            return time_control(identity, **kwargs)

    def max_intersection(self):
        """Get the maximum intersection of all time axes.

        Intersection of user-given netCDF files on disk
        and/or ``time_control`` objects.

        Returns
        -------
        tuple
            Left and right border of maximum time intersection

        Example
        -------
        To get left and right border of maximum time intersection
        of two netCDF files.::

            from pyhomogenize import time_compare

            sinter, einter = time_compare('input1.nc',
                                          'input2.nc').max_intersection()


        """
        start, end = None, None
        for time in self.times:
            if not start:
                start = time[0]
            if not end:
                end = time[-1]
            if time[0] > start:
                start = time[0]
            if time[-1] < end:
                end = time[-1]
            if start > end:
                start, end = None, None
        return start, end

    def select_max_intersection(self, **kwargs):
        """Select maximum intersection time slice.

        Intersection of user-given netCDF files on disk
        and/or ``time_control`` objects.

        Parameters
        ----------
        kwargs
            Optional parameters transferred to function `select_time_range`
            output

        Returns
        -------
        list
            List of user-given netCDF files on disk
            and/or ``time_control`` objects
            cropped to maximum time intersection.

        Example
        -------
        To crop time axes of two netCDF files to maximum time intersection.::

            from pyhomogenize import time_compare

            time_compare('input1.nc',
                         'input2.nc').select_max_intersection(output='output.nc')


        """
        max_intersection = self.max_intersection()
        if max_intersection == (None, None):
            return
        return [
            tco.select_time_range(max_intersection, **kwargs)
            for tco in self.time_control_objects
        ]



"""Unit test package for pyhomogenize.
Test if all necessary modules are available"""

import importlib

import pytest


def _importskip(modname):
    try:
        importlib.import_module(modname)
        has = True
    except ImportError:
        has = False
    func = pytest.mark.skipif(not has, reason=f"requires {modname}")
    return has, func


has_dask, requires_dask = _importskip("dask")
has_xarray, requires_xarray = _importskip("xarray")
has_numpy, requires_numpy = _importskip("numpy")
has_iteration_utilities, requires_iteration_utilities = _importskip(
    "iteration_utilities"
)



import pyhomogenize as pyh

help = """
showdups : Print duplicated timestamps. At first, merge files if needed.
    usage: pyhomogenize showdups -i ifile1 [ifile2 [ifileN]]
"""


def start(args):
    file = pyh.time_control(args.input_files)
    duplicates = file.get_duplicates()
    print("Duplicated time steps: ", duplicates)
    return duplicates



import pyhomogenize as pyh

help = """
showvar : Print variable names. At first, merge files if needed.
    usage: pyhomogenize showvar -i ifile1 [ifile2 [ifileN]]
"""


def start(args):
    file = pyh.netcdf_basics(args.input_files)
    name = file.name
    print(name)
    return name



# -*- coding: utf-8 -*-
# flake8: noqa

import pytest

import pyhomogenize as pyh

from . import (has_dask, has_iteration_utilities, has_numpy, has_xarray,
               requires_dask, requires_iteration_utilities, requires_numpy,
               requires_xarray)

netcdffile = pyh.test_netcdf[0]


def test_netcdf_basics():
    netcdfbasics = pyh.netcdf_basics(netcdffile)
    assert netcdfbasics.files
    assert netcdfbasics.ds
    assert netcdfbasics.name
    assert netcdfbasics.write(output="test.nc")


def test_netcdf_basics_fmt():
    netcdfbasics = pyh.netcdf_basics(netcdffile, fmt="%Y%m%d")
    assert netcdfbasics.fmt



import xarray as xr
from iteration_utilities import duplicates

from . import _consts as consts
from ._netcdf_basics import netcdf_basics


class time_control(netcdf_basics):
    """Class for dealing with a netCDF file's time axis.

    The :class:`time_control` contains the :class:`ǹetcdf_basics`.
    """

    def __init__(self, *args, **kwargs):
        netcdf_basics.__init__(self, *args, **kwargs)
        self.time = self.time()
        del self.frequency
        self.frequency = self.frequency()
        self.time_fmt = self.time_fmt()
        self.equalize = self.equalize()
        del self.calendar
        self.calendar = self.calendar()

    def time(self):
        """netCDF file's time axis"""
        return self._convert_time(self.ds.time)

    def frequency(self):
        """netCDF file's frequency"""
        return self._get_frequency()

    def time_fmt(self):
        """predefined explicit format string derived from `frequency`"""
        return consts.format[self.ds.frequency]

    def equalize(self):
        """predefined list of ``datetime.datetime`` instance attributes
        to be ignored
        """
        return consts.equalize[self.ds.frequency]

    def calendar(self):
        """Calendar type read from netCDF file"""
        return self.time.calendar

    def _get_frequency(self):
        """Get frequency of xr.Dataset"""
        frequency = xr.infer_freq(self.ds.time)
        if not frequency:
            try:
                frequency = consts.frequencies[self.ds.frequency]
            except Exception:
                print("Could not determine any frequency")
                return
        return frequency

    def _duplicates(self):
        """Get duplicated time steps."""
        time = self._equalize_time(self.time, ignore=self.equalize)
        return sorted(list(duplicates(time)))

    def _missings(self):
        """Get missing time steps."""
        time = self._equalize_time(self.time, ignore=self.equalize)
        date_range = self._equalize_time(
            self.date_range(
                time[0], time[-1], self.frequency, calendar=self.time.calendar
            ),
            ignore=self.equalize,
        )
        return sorted(list(set(date_range).difference(time)))

    def _redundants(self):
        """Get redundant time steps."""
        time = self._equalize_time(self.time, ignore=self.equalize)
        date_range = self._equalize_time(
            self.date_range(
                time[0], time[-1], self.frequency, calendar=self.time.calendar
            ),
            ignore=self.equalize,
        )
        return sorted(list(set(time).difference(date_range)))

    def _write_timesteps(self, timesteps, naming):
        """Write timesteps to variable attributes."""
        timesteps = self._convert_to_string(timesteps)
        self._dictionary(naming, self.name, timesteps)
        self.to_variable_attributes(timesteps, naming)

    def get_duplicates(self):
        """Get string of duplicated time steps."""
        return self._convert_to_string(self._duplicates())

    def get_missings(self):
        """Get string of missing time steps."""
        return self._convert_to_string(self._missings())

    def get_redundants(self):
        """Get string of redundant time steps."""
        return self._convert_to_string(self._redundants())

    def check_timestamps(
        self,
        selection=["duplicates", "redundants", "missings"],
        output=None,
        correct=False,
    ):
        """Check netCDF file's time axis.

        Exist whether duplicated, missing and/or redundant time steps.

        Parameters
        ----------
        selection: str or list, default=['duplicates','redundants','missings']
            Check which kind of time steps exist.
        output: str, optional
            Write result on disk.
        correct: bool, default: False
            Delete located time steps from xr.Dataset.
            Automatically set True if output.

        Returns
        -------
        self

        Example
        -------
        To check netCDF file's time axis whether duplicated, missing and/or
        redundant time steps exist and result write on disk::

            from pyhomogenize import time_control

            time_control('input.nc').check_timestamps(output='output.nc')

        """
        if isinstance(selection, str):
            selection = [selection]
        deletes = []
        time = self._equalize_time(self.time, ignore=self.equalize)
        for select in selection:
            nmng = consts.naming[select]
            if not select.startswith("_"):
                select = "_" + select
            add = getattr(self, select)()

            for a in add:
                loc = [i for i, e in enumerate(time) if e == a][1:]
                deletes += loc
            self._write_timesteps(add, nmng)
        dlist = list(dict.fromkeys(deletes))
        timesteps = [n for n, t in enumerate(time) if n not in dlist]
        if output:
            correct = True
        if correct:
            self.ds = self.ds.isel(time=timesteps)
            self.time = self._convert_time(self.ds.time)
        if output:
            self.write(input=self.ds, output=output)
        return self

    def select_time_range(self, time_range, output=None):
        """Select user-given time slice from xr.Dataset

        Parameters
        ----------
        time_range: list
            List of two strings or ``cftime.datatime`` object
            representing the left and right time bounds
        output: str, optional
            Write result on disk.

        Returns
        -------
        self

        Example
        -------
        To select time slice from netCDF file.::

            from pyhomogenize import time_control

            time_control('input.nc').select_time_range(
                                         ['2005-01-01','2005-12-31'],
                                         output='output.nc')
        """
        start_date, end_date = time_range
        if not isinstance(start_date, str):
            start_date = self.date_to_str(start_date)
        if not isinstance(end_date, str):
            end_date = self.date_to_str(end_date)
        self.ds = self.ds.sel(time=slice(start_date, end_date))
        self.time = self._convert_time(self.ds.time)
        if output:
            self.write(input=self.ds, output=output)
        return self

    def select_limited_time_range(self, output=None, **kwargs):
        """Select time slice from xr.Dataset satisfying user-given conditions.
        See pyh.basics.date_range_to_frequency_limits.

        Parameters
        ----------
        output: str, optional
            Write result on disk.
        kwargs:
            Optional parameters transferred to function
            `date_range_to_frequency_limits`:
            smonth
            emonth
            is_month_start
            is_month_end

        Returns
        -------
        self

        Example
        -------
        To select time slice from netCDF file starts
        with the first month of any season and
        ends with the last month of any season.
        The time slice is then e.g. from 2005-03-16 to 2005-11-16::

            from pyhomogenize import time_control

            time_control('input.nc').select_limited_time_range(
                                               smonth=[3,6,9,12],
                                               emonth=[2,5,8,11],
                                               output='output.nc')

        """
        start, end = self.date_range_to_frequency_limits(
            self, date_range=self.time, frequency=self.frequency, **kwargs
        )
        start_date = self.date_to_str(start)
        end_date = self.date_to_str(end)
        self.ds = self.ds.sel(time=slice(start_date, end_date))
        self.time = self._convert_time(self.ds.time)
        if output:
            self.write(input=self.ds, output=output)
        return self

    def within_time_range(self, requested_time_range, fmt=None):
        """
        Checks whether netCDF files time axis is within user-given borders.

        Parameters
        ----------
        requested_time_range: list
            List of two strings or ``cftime.datatime`` object representing
            the left and right time bounds
        fmt: str, default: '%Y-%m-%dT%H:%M:%S'
            Explicit format string for converting string into
            ``cftime.datetime`` object

        Returns
        -------
        bool

        Example
        -------

        To check whether netCDF files time axis is within user-given borders.::

            from pyhomogenize import time_control

            within = time_control('input.nc').within_time_range(['2005-01-02',
                                                                 '2005-12-31'])
        """
        avail_start = self.time[0]
        avail_end = self.time[-1]
        req_start = requested_time_range[0]
        req_end = requested_time_range[-1]
        if isinstance(req_start, str):
            req_start = self.str_to_date(
                req_start,
                fmt=fmt,
                calendar=self.calendar,
            )
        if isinstance(req_end, str):
            req_end = self.str_to_date(
                req_end, fmt=fmt, calendar=self.calendar, mode="end"
            )
        try:
            key = self.ds.frequency
        except Exception:
            key = self._get_key_to_value(consts.frequencies, self.frequency)
        for unit_of_time in consts.within[key]:
            astart = getattr(avail_start, unit_of_time)
            rstart = getattr(req_start, unit_of_time)
            if astart > rstart:
                return False
            if astart == rstart:
                continue
            break
        for unit_of_time in consts.within[key]:
            aend = getattr(avail_end, unit_of_time)
            rend = getattr(req_end, unit_of_time)
            if aend < rend:
                return False
            if aend == rend:
                continue
            break

        return True



import pyhomogenize as pyh

help = """
showreds : Print redundant timestamps. At first, merge files if needed.
    usage: netcdf_time_control showreds ifile1 [ifile2 [ifileN]]
"""


def start(args):
    file = pyh.time_control(args.input_files)
    redundants = file.get_redundants()
    print("Redundant time steps: ", redundants)
    return redundants



import pyhomogenize as pyh

help = """
seltimerange : Select user-given time range. At first, merge files if needed.
    usage: pyhomogenize seltimerange,<timestamp1>,<timestapm2>
           -i ifile1 [ifile2 [ifileN]] -o ofile
    timeformat: %y%m%d[T:%H:%M:%S]
"""


def start(args):
    file = pyh.time_control(args.input_files)
    if not args.output_file:
        print("No output file selecetd. Use -o <ofile>.")
    file.select_time_range(args.arguments, output=args.output_file)
    return file.ds



frequencies = {
    "1hr": "1H",
    "3hr": "3H",
    "6hr": "6H",
    "6hrPt": "6H",
    "day": "D",
    "mon": ["MS", "M"],
    "monClim": ["MS", "M"],
    "yr": ["AS", "A"],
    "fx": None,
}

translator = {
    "1hr": "hour",
    "3hr": "hour",
    "6hr": "hour",
    "6hrPt": "hour",
    "day": "day",
    "mon": "month",
    "monClim": "month",
    "yr": "year",
    "fx": None,
}

format = {
    "1hr": "%Y-%m-%dT%H:%M",
    "3hr": "%Y-%m-%dT%H:%M",
    "6hr": "%Y-%m-%dT%H:%M",
    "6hrPt": "%Y-%m-%dT%H:%M",
    "day": "%Y-%m-%d",
    "mon": "%Y-%m",
    "monClim": "%Y-%m",
    "yr": "%Y",
    "fx": None,
}

is_month = {
    "hour": True,
    "day": True,
    "month": False,
    "year": False,
    None: False,
}

equalize = {
    "1hr": [
        "second",
        "microsecond",
        "nanosecond",
    ],
    "3hr": [
        "second",
        "microsecond",
        "nanosecond",
    ],
    "6hr": [
        "second",
        "microsecond",
        "nanosecond",
    ],
    "6hrPt": [
        "second",
        "microsecond",
        "nanosecond",
    ],
    "day": [
        "hour",
        "minute",
        "second",
        "microsecond",
        "nanosecond",
    ],
    "mon": [
        "day",
        "hour",
        "minute",
        "second",
        "microsecond",
        "nanosecond",
    ],
    "monClim": [
        "day",
        "hour",
        "minute",
        "second",
        "microsecond",
        "nanosecond",
    ],
    "yr": [
        "month",
        "day",
        "hour",
        "minute",
        "second",
        "microsecond",
        "nanosecond",
    ],
}

within = {
    "1hr": NotImplemented,
    "3hr": NotImplemented,
    "6hr": NotImplemented,
    "6hrPt": NotImplemented,
    "day": ["year", "month", "day"],
    "mon": ["year", "month"],
    "monClim": ["year", "month"],
    "yr": ["year"],
    "fx": None,
}

naming = {
    "duplicates": "duplicated_timesteps",
    "redundants": "redundant_timesteps",
    "missings": "missing_timesteps",
}



# -*- coding: utf-8 -*-
# flake8: noqa

"""Top-level package for pyhomogenize."""

import pkg_resources

from . import _consts as consts
from . import _utilities as utilities
from . import operators as op
from ._basics import basics
from ._netcdf_basics import netcdf_basics
from ._time_compare import time_compare
from ._time_control import time_control
from .cli import create_parser
from .data import netcdf as test_netcdf
from .pyhomogenize import pyhomogenize

__author__ = """Ludwig Lierhammer"""
__email__ = "ludwig.lierhammer@hereon.de"
__version__ = "0.1.1"

_all__ = [
    "consts",
    "utilities",
    "op",
    "basics",
    "netcdf_basics",
    "time_compare",
    "time_control",
    "create_parser",
    "test_netcdf",
    "pyhomogenize",
]



# -*- coding: utf-8 -*-
# flake8: noqa

import pytest

import pyhomogenize as pyh

from . import (has_dask, has_numpy, has_xarray, requires_dask, requires_numpy,
               requires_xarray)


def test_time_compare():
    netcdffile1 = pyh.test_netcdf[0]
    netcdffile2 = pyh.test_netcdf[2]
    time_control1 = pyh.time_control(netcdffile1)
    time_control2 = pyh.time_control(netcdffile2)

    assert pyh.time_compare(
        time_control1.ds, time_control2.ds
    ).select_max_intersection()
    assert pyh.time_compare(
        [time_control1.ds, time_control2.ds]
    ).select_max_intersection()
    assert pyh.time_compare(
        [time_control1.ds], time_control2.ds
    ).select_max_intersection(output="test.nc")



import copy
from datetime import datetime as dt
from datetime import timedelta as td

import cftime
import pandas as pd
import xarray as xr

from . import _consts as consts


class basics:
    """Class for controlling netCDF CF standard time axis.

    The :class:`basics` contains some basics functions.

    Parameters
    ----------
    fmt: str, default: '%Y-%m-%dT%H:%M:%S'
        Time format for converting strings into ``cftime.datetime`` object
    calendar: str, default: 'standard'
        Calendar type for the datetimes.
    """

    def __init__(
        self,
        fmt="%Y-%m-%dT%H:%M:%S",
        calendar="standard",
        frequency="D",
    ):
        self.fmt = fmt
        self.calendar = calendar
        self.frequency = frequency

    def fmt(self, fmt):
        """Time format for converting strings into
        ``cftime.datetime`` object.
        """
        return fmt

    def calendar(self, calendar):
        """Calendar type for the datetimes.
        Will be overwritten by netCDF file's inherent calendar.
        """
        return calendar

    def frequency(self, frequency):
        """Frequency strings or list of strings can have multiples.
        Will be overwritten by netCDF file's inherent frequency.
        """
        return frequency

    def _flatten_list(self, lst):
        """Flatten a list containing strings and lists of arbitrarily nested lists

        Parameters
        ----------
        lst: list
            List containing strings and lists of arbitrarily nested lists

        Returns
        -------
        list
        """
        rt = []
        for i in lst:
            if isinstance(i, list):
                rt.extend(self._flatten_list(i))
            else:
                rt.append(i)
        return rt

    def _convert_to_string(self, values, delim=",", fmt=None):
        """Converts list of strings and ``cftime.datetime`` or
        ``datetime.datetime`` objects to string.

        Parameters
        ----------
        values: list
            List of strings and ``cftime.datetime`` or
            ``datetime.datetime`` objects
        delim: str, default: ','
            Output string delimiter between input list entries
        fmt: str, default: '%Y-%m-%dT%H:%M:%S'
            Explicit format string for converting string into
            ``cftime.datetime`` object
            Consider only if list element is ``cftime.datetime`` object

        Returns
        -------
        str
        """
        if not fmt:
            fmt = self.fmt
        converted = ""
        for v in values:
            try:
                v = self.date_to_str(v, fmt=fmt)
            except Exception:
                pass
            converted += str(v) + delim
        return converted[:-1]

    def _dictionary(self, attr, keys, value):
        """Write or update attribute using key-value pairs

        Parameters
        ----------
        attr: str
            Attribute name
        keys: str
            Key names
        values: any type
            Key values
        """
        for key in keys:
            if attr in self.__dict__:
                getattr(self, attr)[key] = value
            else:
                setattr(self, attr, {key: value})

    def _get_key_to_value(self, dict, value):
        """Get key of key-value pair using value

        Parameters
        ----------
        dict: dict
            Dictionary containing key-value pair
        value: any type
            Value of key-value pair

        Returns
        -------
        key: str
            Key of key-value pair
        """
        key_list = list(dict.keys())
        val_list = list(dict.values())
        position = val_list.index(value)
        return key_list[position]

    def _get_date_attr(self, frequencies):
        """Get ``cftime.datetime`` instance attribute from CF frequency

        Parameters
        ----------
        frequencies: str or list
            CF frequency or list of CF frequencies

        Returns
        -------
        attribute: str
            ``cftime.datetime`` instance attribute
        """
        f = self._get_key_to_value(consts.frequencies, frequencies)
        return consts.translator[f]

    def str_to_date(self, str, fmt=None, mode="start", calendar=None):
        """Converts string to ``cftime.datetime`` object

        Parameters
        ----------
        str: str
            string representing the time
        fmt: str, default: '%Y-%m-%dT%H:%M:%S'
            Explicit format string
        calendar: str, default: 'standard'
            Calendar type for the datetimes
        mode: {'start', 'end'}, default: 'start'
            `start`: Set ``cftime.datetime`` instance attributes
            not represented by `fmt` to 0.
            `end`: Set ``cftime.datetime`` instance attributes
            not represented by `fmt` to last possible values
            considering `str`.

        Returns
        -------
        ``cftime.datetime`` object
        """
        i = 1
        if not fmt:
            fmt = self.fmt
        if not calendar:
            calendar = self.calendar
        if mode not in ["start", "end"]:
            return
        while True:
            try:
                date = dt.strptime(str, fmt[:i])
                break
            except Exception:
                i += 1
        cfdate = cftime.datetime(
            date.year,
            date.month,
            date.day,
            calendar=calendar,
        )
        if mode == "start":
            return cfdate
        if mode == "end":
            return cfdate + td(days=1) - td(seconds=1)

    def date_to_str(self, date, fmt=None):
        """Converts ``cftime.datetime`` or ``datetime.datetime`` object to string

        Parameters
        ----------
        date: ``cftime.datetime`` or ``datetime.datetime`` object
            time to be converted
        fmt: str, default: '%Y-%m-%dT%H:%M:%S'
            Explicit format string

        Returns
        -------
        str: str
            string representing the time
        """
        if not fmt:
            fmt = self.fmt
        return date.strftime(fmt)

    def _convert_time(self, time):
        """Converts time object to ``datetime.datetime`` object"""
        try:
            return pd.to_datetime(
                time.indexes["time"].astype("str"),
                format=time.units.split(" ")[-1],
            )
        except Exception:
            return time.indexes["time"]

    def _equalize_time(
        self,
        time,
        ignore=["second", "microsecond", "nanosecond"],
    ):
        """Ignore ``datetime.datetime`` instance attributes and set them to 0

        Parameters
        ----------
        time: CFTimeIndex
            CFTimeIndex containing cftime.datetime objects
        ignore: list, default: ['second','microsecond','nanosecond']
            list of datetime.datetime instance attributes to be ignored

        Returns
        -------
        time: CFTimeIndex
            ``CFTimeIndex`` containing ``cftime.datetime`` objects
            with ignored instance attributes
        """
        n = 0
        while n < len(ignore):
            etime = time.tolist()
            query = {ignr: 1 for ignr in ignore[::-1][n:]}
            try:
                i = 0
                while i < len(etime):
                    etime[i] = etime[i].replace(**query)
                    i += 1
                return etime
            except Exception:
                n += 1
        return time.tolist()

    def _interpret_frequency(self, freq):
        if isinstance(freq, str):
            if freq in consts.frequencies.keys():
                freq = consts.frequencies[freq]
        return freq

    def _mid_timestep(self, freq, st, end, calendar=None):
        """Build ``CFTimeIndex``
        Set elements between user-given frequencies

        Parameters
        ----------
        freq: list
            list of two frequency strings for use with ``cftime`` calendars
            https://xarray.pydata.org/en/stable/generated/xarray.cftime_range.html
        st: cftime.datetime or datetime.datetime
            Left bound for generating dates
        end: cftime.datetime or datetime.datetime
            Right bound for generating dates
        calendar: str, default: 'standard'
            Calendar type for the datetimes

        Returns
        -------
        CFTimeIndex
        """

        def end_of(date, attr):
            while True:
                m = getattr(date, attr)
                nm = getattr((date + td(hours=1)), attr)
                if nm != m:
                    return date
                date += td(hours=1)

        def begin_of(date, attr):
            while True:
                m = getattr(date, attr)
                pm = getattr((date - td(hours=1)), attr)
                if pm != m:
                    return date
                date -= td(hours=1)

        if not calendar:
            calendar = self.calendar
        t1 = xr.cftime_range(
            st,
            end,
            freq=freq[0],
            calendar=calendar,
        )
        t2 = xr.cftime_range(
            st,
            periods=len(t1),
            freq=freq[1],
            calendar=calendar,
        )
        return t1 + (t2 - t1 + td(days=1)) / 2

    def _point_timestep(self, freq, st, end, calendar=None):
        """Build ``CFTimeIndex``
        Set elements to user-given frequency

        Parameters
        ----------
        freq: str
            frequency string for use with ``cftime`` calendars
            https://xarray.pydata.org/en/stable/generated/xarray.cftime_range.html
        st: cftime.datetime or datetime.datetime
            Left bound for generating dates
        end: cftime.datetime or datetime.datetime
            Right bound for generating dates
        calendar: str, default: 'standard'
            Calendar type for the datetimes

        Returns
        -------
        CFTimeIndex
        """
        if not calendar:
            calendar = self.calendar
        return xr.cftime_range(st, end, freq=freq, calendar=calendar)

    def date_range(self, start, end, frequency="D", calendar=None):
        """Build ``CFTimeIndex``

        Parameters
        ----------
        start: cftime.datetime or datetime.datetime or str
            Left bound for generating dates
        end: cftime.datetime or datetime.datetime or str
            Right bound for generating dates
        frequency: str ot list
            frequency string or list of frequency strings
            for use with ``cftime`` calendars
            https://xarray.pydata.org/en/stable/generated/xarray.cftime_range.html
            Set elements to user-given frequency if type of frequency is str
            Set elements between user-given frequencies if type
        calendar: str, default: 'standard'

        Returns
        -------
        CFTimeIndex
        """
        if not calendar:
            calendar = self.calendar
        frequency = self._interpret_frequency(frequency)
        if isinstance(frequency, str):
            return self._point_timestep(
                frequency,
                start,
                end,
                calendar=calendar,
            )
        if isinstance(frequency, list):
            return self._mid_timestep(
                frequency,
                start,
                end,
                calendar=calendar,
            )

    def is_month_start(self, cftime_range):
        """Check whether each element of ``CFTimeIndex`` is first day of the month

        Parameters
        ----------
        cftime_range: CFTimeIndex
            CFTimeIndex containing cftime.datetime objects

        Returns
        -------
        list
            list of boolean values
        """
        array = []
        for cftr in cftime_range:
            prev = cftr - td(days=1)
            if prev.month != cftr.month:
                array += [True]
                continue
            array += [False]
        return array

    def is_month_end(self, cftime_range):
        """Check whether each element of ``CFTimeIndex`` is last day of the month

        Parameters
        ----------
        cftime_range: CFTimeIndex
            CFTimeIndex containing cftime.datetime objects

        Returns
        -------
        list
            list of boolean values
        """

        array = []
        for cftr in cftime_range:
            next = cftr + td(days=1)
            if next.month != cftr.month:
                array += [True]
                continue
            array += [False]
        return array

    def date_range_to_frequency_limits(
        self,
        start=None,
        end=None,
        date_range=None,
        frequency=None,
        calendar=None,
        smonth=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
        emonth=[12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1],
        is_month_start=None,
        is_month_end=None,
        get_range=False,
    ):
        """Get bounds of CFTimeIndex which satisfy user-given conditions.

        Parameters
        ----------
        start: str or datetime.datetime or cftime.cftime, optional
            Left bound for generating dates
        end: str or datetime.datetime or cftime.cftime, optional
            Right bound for generating dates
        date_range: CFTimeIndex
            CFTimeIndex containing cftime.datetime objects
        frequency: str ot list, default:'D'
            CF frequency string or list of CF frequency strings
            or frequency string or list of frequency strings
            for use with ``cftime`` calendars
            https://xarray.pydata.org/en/stable/generated/xarray.cftime_range.html
        calendar: str, default: 'standard'
            Calendar type for the datetimes
        smonth: list, default: [1,2,3,4,5,6,7,8,9,10,11,12]
            One of the allowed values of left bound's
            ``datetime.datetime`` instance attribute month
        emonth: list, default: [12,11,10,9,8,7,6,5,4,3,2,1]
            One of the allowed values of right bound's
            ``datetime.datetime`` instance attribute month
        is_month_start: bool, optional
            Value of left bound's ``datetime.datetime`` instance attribute day
            must be equal to first day of the month
            Automatically set to True for sub_monthly frequencies
        is_month_end: bool, optional
            Value of right bound's ``datetime.datetime`` instance attribute day
            must be equal to last day of the month
            Automatically set to True for sub_monthly frequencies
        get_range: bool, default:False
            If False returns left and right bounds
            If True returns CFTimeIndex

        Returns
        -------
        tuple
            Tuple of start and end dates satisfying all conditions

        Example
        -------
        To get bounds of CFTimeIndex which satisfy user-given conditions.
        Monthly frequency values with a 360-day calendar have to start with
        the first month of any season and have to end with the last month
        of any season.
        It returns the tuple (2005-03-16 00:00:00 2005-11-16 00:00:00)::

            from pyhomogenize import basics

            basics = basics()

            start, end = basics.date_range_to_frequency_limits(
                     start='2005-01-01',
                     end='2005-12-30',
                     smonth=[3,6,9,12],
                     emonth=[2,5,8,11],
                     calendar='360_day',
            )

        """
        if date_range is None:
            if not frequency:
                frequency = self.frequency
            if not calendar:
                calendar = self.calendar
            frequency = self._interpret_frequency(frequency)
            if start is None or end is None:
                return None, None
            date_range = self.date_range(
                start, end, frequency=frequency, calendar=calendar
            )
        start = date_range[0]
        end = date_range[-1]
        sdate_range = copy.copy(date_range)
        edate_range = copy.copy(date_range)
        if is_month_start is None:
            is_month_start = consts.is_month[self._get_date_attr(frequency)]
        if is_month_end is None:
            is_month_start = consts.is_month[self._get_date_attr(frequency)]
        if is_month_start:
            sdate_range = date_range[self.is_month_start(date_range)]
        if is_month_end:
            edate_range = date_range[self.is_month_end(date_range)]
        if sdate_range.empty:
            return None, None
        if edate_range.empty:
            return None, None
        for sdate in sdate_range:
            if sdate < start:
                continue
            if sdate.month in smonth:
                break
            sdate = None
        if not sdate:
            return None, None
        for edate in reversed(edate_range):
            if edate > end:
                continue
            if edate.month in emonth:
                break
            edate = None
        if not edate:
            return None, None
        if get_range:
            return xr.CFTimeIndex(
                [cft for cft in date_range if cft >= sdate and cft <= edate]
            )
        else:
            return sdate, edate



"""Console script for netcdf_time_control."""
import argparse
import sys

from .pyhomogenize import pyhomogenize


def csv_list(string):
    return string.split(",")


def create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "operator",
        nargs="?",
        default=None,
        help="Operator",
        type=csv_list,
    ),
    parser.add_argument(
        "-i",
        "--input_files",
        dest="input_files",
        nargs="+",
        help="List of input files",
    )
    parser.add_argument(
        "-o",
        "--output_file",
        dest="output_file",
        help="Name of the putput file",
    )
    parser.add_argument(
        "-ops",
        "--operators",
        dest="operators",
        action="store_true",
        help="Choose to get a list of all available operators.",
    )
    return parser


def main():
    """Console script for netcdf_time_control."""
    parser = create_parser()
    args = parser.parse_args()
    pyhomogenize(args)
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



# -*- coding: utf-8 -*-
# flake8: noqa

import pytest

import pyhomogenize as pyh

from . import (has_dask, has_iteration_utilities, has_numpy, has_xarray,
               requires_dask, requires_iteration_utilities, requires_numpy,
               requires_xarray)

netcdffile = [pyh.test_netcdf[1], pyh.test_netcdf[3]]
time_control = pyh.time_control(netcdffile)


def test_get_duplicates():
    time_control.get_duplicates()


def test_get_missings():
    time_control.get_missings()


def test_get_redundants():
    time_control.get_redundants()


def test_check_timestamps():
    assert time_control.check_timestamps(output="test.nc")
    assert time_control.check_timestamps(selection="duplicates")
    assert time_control.check_timestamps(selection=["duplicates", "missings"])


def test_select_time_range():
    assert time_control.select_time_range(
        ["2007-06-01", "2008-06-30"], output="test.nc"
    )


def test_within_time_range():
    time_control.within_time_range(["2007-06-01", "2008-06-30"])
    time_control.within_time_range(["2006-06-01", "2008-06-30"])
    time_control.within_time_range(["20070601", "20080630"], fmt="%Y%m%d")


def test_select_limited_time_range():
    time_control.select_limited_time_range(
        output="test.nc", smonth=[3, 6, 9, 12], emonth=[2, 5, 8, 11]
    )



import pyhomogenize as pyh

help = """
merge : Merge given input files
    usage: pyhomogenize merge -i ifile1 [ifile2 [ifileN]] -o ofile
"""


def start(args):
    file = pyh.netcdf_basics(args.input_files)
    if not args.output_file:
        print("No output file selecetd. Use -o <ofile>.")
    else:
        file.write(output=args.output_file)
    return file.ds



import pyhomogenize as pyh

help = """
timecheck : By default, delete duplicated and redundant time stamps
from input files and write duplicated, redundant and missing timestamps
to netcdf variable attributes. The selection is changeable.
    usage: pyhomogenize timecheck[,duplicates,redundants,missings]
                        -i ifile1 [ifile2 [ifileN]] -o ofile
"""


def start(args):
    file = pyh.time_control(args.input_files)
    if not args.output_file:
        print("No output file selecetd. Use -o <ofile>.")
    if args.arguments:
        file.check_timestamps(
            selection=args.arguments,
            output=args.output_file,
        )
    else:
        file.check_timestamps(output=args.output_file)
    if not args.output_file:
        if hasattr(file, "duplicated_timesteps"):
            print("Duplicated time steps: ", file.duplicated_timesteps)
        if hasattr(file, "redundant_timesteps"):
            print("Redundant time steps: ", file.redundant_timesteps)
        if hasattr(file, "missing_timesteps"):
            print("Missing time steps: ", file.missing_timesteps)
    return file.ds



# -*- coding: utf-8 -*-
# flake8: noqa

import pytest

import pyhomogenize as pyh

from . import (has_dask, has_iteration_utilities, has_numpy, has_xarray,
               requires_dask, requires_iteration_utilities, requires_numpy,
               requires_xarray)


def test_cli_ops():
    parser = pyh.create_parser()
    args = parser.parse_args(["-ops"])
    pyh.pyhomogenize(args)


def test_cli_merge():
    parser = pyh.create_parser()
    args = parser.parse_args(["merge", "-i", pyh.test_netcdf[1], pyh.test_netcdf[3]])
    assert pyh.pyhomogenize(args)


def test_cli_seltimerange():
    parser = pyh.create_parser()
    args = parser.parse_args(
        ["seltimerange,20070501,20070630", "-i", pyh.test_netcdf[1]]
    )
    assert pyh.pyhomogenize(args)


def test_cli_showdups():
    parser = pyh.create_parser()
    args = parser.parse_args(["showdups", "-i", pyh.test_netcdf[0]])
    pyh.pyhomogenize(args)


def test_cli_showmiss():
    parser = pyh.create_parser()
    args = parser.parse_args(["showmiss", "-i", pyh.test_netcdf[0]])
    pyh.pyhomogenize(args)


def test_cli_showreds():
    parser = pyh.create_parser()
    args = parser.parse_args(["showreds", "-i", pyh.test_netcdf[0]])
    pyh.pyhomogenize(args)


def test_cli_showtimestamps():
    parser = pyh.create_parser()
    args = parser.parse_args(["showtimestamps", "-i", pyh.test_netcdf[0]])
    pyh.pyhomogenize(args)


def test_cli_showvar():
    parser = pyh.create_parser()
    args = parser.parse_args(["showvar", "-i", pyh.test_netcdf[0]])
    pyh.pyhomogenize(args)


def test_cli_timecheck_all():
    parser = pyh.create_parser()
    args = parser.parse_args(["timecheck", "-i", pyh.test_netcdf[0]])
    assert pyh.pyhomogenize(args)


def test_cli_timecheck_sel():
    parser = pyh.create_parser()
    args = parser.parse_args(["timecheck,duplicates", "-i", pyh.test_netcdf[0]])
    assert pyh.pyhomogenize(args)



import pyhomogenize as pyh

help = """
showmiss : Print missing timestamps. At first, merge files if needed.
    usage: pyhomogenize showmiss ifile1 [ifile2 [ifileN]]
"""


def start(args):
    file = pyh.time_control(args.input_files)
    missings = file.get_missings()
    print("Missing time steps: ", missings)
    return missings



import pyhomogenize as pyh

help = """
showtimestamps : Show available timestamps. At first, merge files if needed.
    usage: pyhomogenize showtimestamps -i ifile1 [ifile2 [ifileN]]
"""


def start(args):
    file = pyh.time_control(args.input_files)
    time = file.time
    print(time)
    return time



from . import (merge, seltimerange, showdups, showmiss, showreds,
               showtimestamps, showvar, timecheck)

operators = """

All available operators implemented:
Arguments in brackets are optional.
"""

operators += merge.help
operators += showvar.help
operators += seltimerange.help
operators += showtimestamps.help
operators += showdups.help
operators += showmiss.help
operators += showreds.help
operators += timecheck.help



import xarray as xr

from ._basics import basics


class netcdf_basics(basics):
    """Class for reading an writing netCDF files.

    The :class:`netcdf_basics` contains :class:`basics`.

    Parameters
    ----------
    files: str or list
        file on disk or xarray.Dataset or list of both
    """

    def __init__(self, files, **kwargs):
        basics.__init__(self, **kwargs)
        if isinstance(files, str):
            files = [files]
        self.files = self.files(files)
        self.ds = self.ds()
        self.name = self.name()

    def files(self, files):
        """List of input netCDF file(s) and/or xr.Dataset(s)."""
        return files

    def ds(self):
        """Input netCDF file(s) openend as xr.Dataset(s). See method open"""
        return self.open()

    def name(self):
        """CF variable name of `ds`. See method get_var_name"""
        return self.get_var_name()

    def _add_to_attrs(self, target, attr_name, value):
        """Adds or updates attribute

        Parameters
        ----------
        target: xr.Dataset or xr.DataArray
            target of adding or updating attribute
        attr_name: str
            Name of the attribute which will be added or updated
        value: str
            Name of the value to be added
        """
        if attr_name in target.attrs:
            value = target.attrs[attr_name] + ", " + value
        target.attrs[attr_name] = value

    def to_variable_attributes(self, indexes, attr_name):
        """Adds or updates variable attributes

        Parameters
        ----------
        indexes: str or list
            Attribute value to be added
        attr_name: str
            Name of the attribute which will be added or updated
        """
        var_name = self.name
        if isinstance(indexes, list):
            indexes = str(indexes)
        if indexes:
            for var in var_name:
                self._add_to_attrs(getattr(self.ds, var), attr_name, indexes)

    def to_global_attributes(self, indexes, attr_name):
        """Adds or updates gloabl attributes

        Parameters
        ----------
        indexes: str or list
            Attribute value to be added
        attr_name: str
            Name of the attribute which will be added or updated
        """
        if isinstance(indexes, list):
            indexes = str(indexes)
        if indexes:
            self._add_to_attrs(self.ds, attr_name, indexes)

    def get_var_name(self, ds=None):
        """List of CF variables in xr.Dataset

        Parameters
        ----------
        ds: xr.Dataset
            xarray Dataset
        Returns
        -------
        list
            List of CF variables
        """

        def condition(ds, var):
            return len(ds[var].coords) == len(ds.coords)

        def most_coords(ds):
            coords = 0
            name = []
            for var in ds.data_vars:
                ncoords = len(ds[var].coords)
                if ncoords > coords:
                    coords = ncoords
                    name += [var]
            return name

        if not ds:
            ds = self.ds

        try:
            var_list = [
                var
                for var in self.ds.data_vars
                if condition(
                    self.ds,
                    var,
                )
            ]
            if var_list:
                return var_list
            return most_coords(ds)
        except Exception:
            return [self.ds.name]

    def _open_xrdataset(
        self,
        files,
        use_cftime=True,
        parallel=True,
        data_vars="minimal",
        chunks={"time": 1},
        coords="minimal",
        compat="override",
        drop=None,
        **kwargs
    ):
        """optimized function for opening large cf datasets.
        based on
        https://github.com/pydata/xarray/issues/1385#issuecomment-561920115
        decode_timedelta=False is added to leave variables and
        coordinates with time units in
        {“days”, “hours”, “minutes”, “seconds”, “milliseconds”, “microseconds”}
        encoded as numbers.

        """

        def drop_all_coords(ds):
            return ds.reset_coords(drop=True)

        ds = xr.open_mfdataset(
            files,
            parallel=parallel,
            decode_times=False,
            combine="by_coords",
            preprocess=drop_all_coords,
            decode_cf=False,
            chunks=chunks,
            data_vars=data_vars,
            coords=coords,
            compat=compat,
            **kwargs
        )

        return xr.decode_cf(ds, use_cftime=use_cftime, decode_timedelta=False)

    def open(self):
        """Opens file or list of files on disk.
        Result is automaticaly wrote to object's attributes.
        """
        if isinstance(self.files, xr.Dataset):
            return self.files
        elif isinstance(self.files, str):
            return self._open_xrdataset(self.files)
        elif isinstance(self.files, list):
            if all(isinstance(x, (xr.Dataset)) for x in self.files):
                return xr.concat(self.files, dim="time")
            elif all(isinstance(x, (str)) for x in self.files):
                return self._open_xrdataset(self.files)
        raise ValueError(
            "Input files are not xarray Datasets or files on disk."
            "You can not mix those two types."
        )

    def write(self, input=None, output=None):
        """Writes `self.ds` or user-given xr.Dataset as netCDF file on disk.

        Parameters
        ----------
        input: xr.Dataset, optional
            xr.Dataset to be written on disk
        output: str, optional
            Name of the output netCDF file

        Returns
        -------
        self
        """

        def is_dataset(input):
            if not isinstance(input, xr.Dataset):
                input = self.ds
            return input

        input = is_dataset(input)
        self.to_variable_attributes(
            self._convert_to_string(self.files), "associated_files"
        )
        if not output:
            print("No output selected.")
        else:
            input.to_netcdf(output)
        return self



import os


def check_existance(files):
    """
    Check if requested files are available
    Exit if not.
    """
    stop = False
    commands = ""
    if not files:
        print("No input files selected.")
        return
    for file in files:
        if not os.path.isfile(file):
            stop = True
            commands += "{} is not available\n".format(file)
    if stop:
        print(commands)
        return
    return True


def get_operator(object, name, type="attribute"):
    if not name:
        print("No {} is selected.".format(type))
        return
    try:
        return getattr(object, name)
    except Exception:
        print("Choosen {} {} is not available.".format(type, name))
        return



#!/usr/bin/env python

"""The setup script."""

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

# with open("pyhomogenize/__init__.py") as init_file:
#    lines = init_file.read().strip().replace(" ", "").split("\n")
#    for line in lines:
#        if "__version__" in line:
#            __version__ = line.split("=")[-1]
#            break


def _read_txt(txt_file):
    return open(txt_file).read().strip().split("\n")


requirements = _read_txt("ci/requirements/requirements.txt")

setup_requirements = _read_txt("ci/requirements/requirements_dev.txt")

test_requirements = []

setup(
    author="Ludwig Lierhammer",
    author_email="ludwig.lierhammer@hereon.de",
    python_requires=">=3.5",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Tool to homogenize netCDF files to CF standard",
    entry_points={
        "console_scripts": [
            "pyhomogenize=pyhomogenize.cli:main",
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="pyhomogenize",
    name="pyhomogenize",
    packages=find_packages(include=["pyhomogenize", "pyhomogenize.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/ludwiglierhammer/pyhomogenize",
    version="0.1.1",
    zip_safe=False,
)



# -*- coding: utf-8 -*-
# flake8: noqa

import pytest
import xarray as xr

import pyhomogenize as pyh

from . import has_xarray, requires_xarray


def test_date_range_to_frequency_limits():
    basics = pyh.basics()
    start = "2005-03-02"
    end = "2005-12-25"
    smonth = [3, 6, 9, 12]
    emonth = [2, 5, 8, 11]

    frequency = "day"
    assert basics.date_range_to_frequency_limits(
        start, end, frequency=frequency, smonth=smonth, emonth=emonth
    )

    date_range = xr.cftime_range(start, end)
    assert basics.date_range_to_frequency_limits(date_range=date_range)

    frequency = "mon"
    assert basics.date_range_to_frequency_limits(
        start, end, frequency=frequency, smonth=smonth, emonth=emonth
    )

    assert basics.date_range_to_frequency_limits(
        start, end, frequency=frequency, is_month_start=True, is_month_end=True
    )

    frequency = ["MS", "M"]
    assert basics.date_range_to_frequency_limits(start, end, frequency=frequency)



# -*- coding: utf-8 -*-
# flake8: noqa

import pytest

import pyhomogenize as pyh


def _get_keys(dictionary):
    return getattr(pyh.consts, dictionary).keys()


def _get_values(dictionary):
    return getattr(pyh.consts, dictionary).values()


def test_frequencies():
    assert _get_keys("frequencies")
    assert _get_values("frequencies")


def test_translator():
    assert _get_keys("translator")
    assert _get_values("translator")


def test_format():
    assert _get_keys("translator")
    assert _get_values("translator")


def test_equalize():
    assert _get_keys("equalize")
    assert _get_values("equalize")


def test_within():
    assert _get_keys("within")
    assert _get_values("within")


def test_naming():
    assert _get_keys("naming")
    assert _get_values("naming")


def test_is_month():
    assert _get_keys("is_month")
    assert _get_values("is_month")



"""Main module."""

from . import _utilities as ut
from . import operators as op


def pyhomogenize(args):
    # show all available operators
    if args.operators:
        print(op.operators)
        return
    operator = args.operator[0]
    arguments = args.operator[1:]
    args.operator = operator
    args.arguments = arguments
    # check if selected operator is available
    func = ut.get_operator(op, args.operator, type="operator")
    if not func:
        return
    # check if all input files are available
    if not ut.check_existance(args.input_files):
        return
    # start operator
    return func.start(args)



# -*- coding: utf-8 -*-
# flake8: noqa

import pytest

import pyhomogenize as pyh


def test_check_existance():
    pyh.utilities.check_existance([])
    pyh.utilities.check_existance(pyh.test_netcdf)
    pyh.utilities.check_existance(["test_netcdf"])
    pyh.utilities.check_existance(pyh.test_netcdf + ["test_netcdf"])


def test_get_operator_none():
    pyh.utilities.get_operator(pyh.op, "", type="operator")


def test_get_operator_false():
    pyh.utilities.get_operator(pyh.op, "test", type="operator")


def test_get_operator_merge():
    assert pyh.utilities.get_operator(pyh.op, "merge", type="operator")


def test_get_operator_showvar():
    assert pyh.utilities.get_operator(pyh.op, "showvar", type="operator")


def test_get_operator_seltimerange():
    assert pyh.utilities.get_operator(pyh.op, "seltimerange", type="operator")


def test_get_operator_showtimestamps():
    assert pyh.utilities.get_operator(pyh.op, "showtimestamps", type="operator")


def test_get_operator_showdups():
    assert pyh.utilities.get_operator(pyh.op, "showdups", type="operator")


def test_get_operator_showmiss():
    assert pyh.utilities.get_operator(pyh.op, "showmiss", type="operator")


def test_get_operator_showreds():
    assert pyh.utilities.get_operator(pyh.op, "showreds", type="operator")


def test_get_operator_timecheck():
    assert pyh.utilities.get_operator(pyh.op, "timecheck", type="operator")
